#!/usr/bin/env python
# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
""" Script generating mutual information and single orbital entropy plots."""

import argparse
import pathlib

import numpy as np
import matplotlib.colors as colors
import matplotlib.cm as cm
import matplotlib.path as mpath
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
from matplotlib.ticker import NullFormatter
from matplotlib.offsetbox import AnnotationBbox, OffsetImage

try:
    from matplotlib._png import read_png
except ModuleNotFoundError:
    from matplotlib.image import imread as read_png

from pybest import __version__
from pybest import filemanager
from pybest.exceptions import ArgumentError


# pylint: disable=too-many-arguments,too-many-locals
def ploti12(
    i12ind1,
    i12ind2,
    i12val,
    orbinit,
    orbfinal,
    thresh,
    s1index,
    s1value,
    indices,
    zoom=0.09,
):
    """Plot mutual information circular plot."""

    plt.rcParams.update(
        {
            "text.usetex": 1,
            # line below turns the ticks to bold
            "text.latex.preamble": [r"\boldmath"],
        }
    )

    norb = orbfinal - orbinit + 1
    orbitals = np.arange(orbinit, orbfinal + 1)
    theta = 2 * np.pi * (orbitals - orbinit) / (norb)
    max_r = 150.0
    radius = np.ones(norb, int) * (max_r + 50.0)
    r = (max_r) * np.ones(norb, int) - 15.00 * ((orbitals - orbinit) % 2)

    plt.figure(figsize=(10, 10))
    #
    # First subplot: mutual information
    #
    ax = plt.subplot(111, polar=True)
    ax.grid(False)
    ax.set_theta_zero_location("N")
    ax.set_theta_direction(-1)
    ax.spines["polar"].set_visible(False)

    for i in range(len(orbitals)):
        offset = -10
        if i % 2 == 0:
            offset = 10
        ax.annotate(
            indices[i],
            # i+1+orbinit,
            xy=(theta[i], r[i] + offset),
            xytext=(0, 0),
            textcoords="offset points",  # ha = 'center', va = 'bottom',
            fontsize=10,
            fontweight="bold",
            horizontalalignment="center",
            verticalalignment="center",
        )
        #
        # Plot orbitals if available.
        #

        filename = f"mo_{str(indices[i])}.png"
        pad_value = -0.05
        try:
            offset = 50
            if i % 2 == 0:
                offset = 70
            arr_hand = read_png(filename)
            imagebox = OffsetImage(arr_hand, zoom=zoom)

            abox = AnnotationBbox(
                imagebox,
                (theta[i], r[i] + offset - 20),
                xycoords="data",
                frameon=False,
            )
            ax.add_artist(abox)
            pad_value = 0.00
        except Exception:
            continue

    ax.yaxis.set_data_interval(0, 22.5)
    ax.xaxis.set_major_formatter(NullFormatter())
    ax.yaxis.set_major_formatter(NullFormatter())

    ax.plot(
        [theta[0], theta[0]],
        [radius[0], radius[0]],
        "",
        lw=2,
        color="w",
        zorder=0,
    )

    #
    # Sort i12 values
    #
    i12 = list(zip(i12val, i12ind1, i12ind2))
    i12.sort(key=lambda x: x[0])

    #
    # Define style and scale
    #
    color_map = cm.get_cmap("CMRmap_r", 256)
    log_scale = colors.LogNorm(vmin=thresh, vmax=2 * np.log(2))

    for val, ind1, ind2 in i12:
        ind_a = np.where(indices == ind1 + 1)
        ind_b = np.where(indices == ind2 + 1)
        if val >= thresh:
            color = color_map(int(256 * log_scale(val / 2 / np.log(2))))
            linewidth = 4 * val / np.log(2)
        else:
            continue
        if linewidth < 1:
            linewidth = 1
        #
        # Plot quadratic Bezier curve
        #
        Path = mpath.Path
        start = (theta[ind_a], r[ind_a])
        end = (theta[ind_b], r[ind_b])
        bezier_curve = mpatches.PathPatch(
            Path(
                [start, (0, 0), end], [Path.MOVETO, Path.CURVE3, Path.CURVE3]
            ),
            fc="none",
            lw=linewidth,
            color=color,
            transform=ax.transData,
        )
        ax.add_patch(bezier_curve)
    #
    # i_12 colormap legend
    #
    smc = plt.cm.ScalarMappable(cmap=color_map, norm=log_scale)
    plt.colorbar(smc, shrink=0.5, pad=pad_value)

    #
    # Define size of circles in mutual information plot
    #
    ind_list = np.array([], dtype=int)
    for ind in indices:
        ind_list = np.append(ind_list, np.where(s1index == ind))
    areavalue = s1value[ind_list]
    colorvalue = areavalue
    areavalue *= 200

    #
    # Define own colormap:
    #
    endcolor = "#191970"
    midcolor2 = "#ff0000"
    midcolor1 = "#228b22"
    cmap2 = colors.LinearSegmentedColormap.from_list(
        "own2", [midcolor1, midcolor2, endcolor]
    )
    cm.register_cmap(cmap=cmap2)

    #
    # Plot circles of different size and color
    #
    ax.scatter(
        theta,
        r,
        s=areavalue,
        alpha=0.75,
        c=colorvalue,
        cmap=cmap2,
        vmin=0.001,
        vmax=1,
        norm=colors.LogNorm(),
        zorder=4,
    )

    ax.set_ylim(0, 225)

    plt.savefig(f"{filemanager.result_path('i12.pdf')}", format="pdf", dpi=300)


def plots1(orbinit, orbfinal, s1index, s1value, smax=0.71):
    """Plot single orbital entropy."""
    plt.figure(figsize=(10, 5))
    ax2 = plt.subplot(111)
    ax2.axis([orbinit - 1, orbfinal + 1, 0, smax])
    ax2.vlines(s1index, [0], s1value, color="r", linewidth=2, linestyle="-")
    plt.ylabel("single-orbital entropy")
    plt.xlabel("Orbital index")
    plt.plot(s1index, s1value, "ro", markersize=8)

    plt.savefig(f"{filemanager.result_path('s1.pdf')}", format="pdf", dpi=300)


def read_i12_data(orbinit, orbfinal, fname):
    """Read mutual information data from file."""
    index1 = np.array([], dtype=int)
    index2 = np.array([], dtype=int)
    value = np.array([])
    with open(fname) as file_in:
        for line in file_in:
            words = line.split()
            if (
                int(words[0]) >= orbinit
                and int(words[1]) >= orbinit
                and int(words[0]) <= orbfinal
                and int(words[1]) <= orbfinal
            ):
                index1 = np.append(index1, int(words[0]) - 1)
                index2 = np.append(index2, int(words[1]) - 1)
                value = np.append(value, float(words[2]))
    return index1, index2, value


def read_s1_data(orbinit, orbfinal, filename):
    """Read single orbital entropies from file."""
    index = np.array([], dtype=int)
    value = np.array([])
    with open(filename) as file_in:
        for line in file_in:
            words = line.split()
            if len(words) != 2:
                raise ArgumentError(
                    f"Expecting 2 fields on each data line in {filename}"
                )
            index = np.append(index, int(words[0]))
            value = np.append(value, float(words[1]))
    if orbfinal:
        newind = np.where(index >= (orbinit))
        index = index[newind]
        value = value[newind]
        newind2 = np.where(index <= orbfinal)
        index = index[newind2]
        value = value[newind2]
    return index, value


def parse_args():
    """Parse arguments."""
    parser = argparse.ArgumentParser(
        prog="pybest-entanglement.py",
        description="This script makes an orbital entanglement plot. It "
        "assumes that the files s1.dat and i12.dat are present in "
        "the current directory. These two files will be used to "
        "create the figure orbital_entanglement.png.",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"{parser.prog} (PyBEST version) {__version__}",
    )

    parser.add_argument(
        "threshold",
        type=float,
        help="Orbitals with a mutual information below this threshold will not "
        "be connected by a line.",
    )
    parser.add_argument(
        "-i",
        default=1,
        type=int,
        nargs="?",
        help="The first orbital to be used for the plot. [default=%(default)s]",
    )
    parser.add_argument(
        "-z",
        default=0.03,
        type=float,
        nargs="?",
        help="Zoom factor for orbitals. [default=%(default)s]",
    )
    parser.add_argument(
        "-f",
        default=None,
        type=int,
        nargs="?",
        help="The last orbital to be used for the plot (inclusive). "
        "[default=last orbital]",
    )
    parser.add_argument(
        "-t",
        default="all",
        type=str,
        nargs="?",
        help="Type of graph to be plotted. If not specified (None) both graphs \
              will be generated. Possible values are None, i (mutual information), \
              s (single-orbital entropy) [default=%(default)s]",
    )
    parser.add_argument(
        "--smax",
        default=0.71,
        type=float,
        nargs="?",
        help="The maximum of s1 scale. [default=%(default)s]",
    )
    parser.add_argument(
        "--iname",
        default="i12.dat",
        type=str,
        nargs="?",
        help="File name to be read in for Iij. [default=%(default)s]",
    )
    parser.add_argument(
        "--sname",
        default="s1.dat",
        type=str,
        nargs="?",
        help="File name to be read in for S1. [default=%(default)s]",
    )
    parser.add_argument(
        "--order",
        default=None,
        metavar="i",
        type=int,
        nargs="+",
        help="The order of the indices to be shown on each graph. [default=%(default)s]",
    )

    return parser.parse_args()


def main():
    """This is main() of CLI app, the flow is following: the data and command
    line arguments are read and turned into pretty pictures in pdf format.
    """
    args = parse_args()

    orbinit = args.i
    orbfinal = args.f
    smax = args.smax
    select = args.t
    zoom = args.z
    iname = pathlib.Path(args.iname)
    sname = pathlib.Path(args.sname)
    order = args.order

    # Read s1.dat and store data
    s1index, s1value = read_s1_data(orbinit, orbfinal, sname)

    if orbfinal is None:
        orbfinal = s1index[-1]

    # Define different order for I12 plot
    if order is None:
        order = np.arange(orbinit, orbfinal + 1)
    if len(order) != (orbfinal - orbinit + 1):
        raise ArgumentError("Order does not have proper length.")

    # Plot only selected graphs
    if select not in ["i", "s", "all"]:
        raise ValueError(f"Unknown type {select}")
    if select in ["s", "all"]:
        plots1(orbinit, orbfinal, s1index, s1value, smax)
    if select in ["i", "all"]:
        # Read i12.dat and store data
        i12index1, i12index2, i12value = read_i12_data(
            orbinit, orbfinal, iname
        )

        ploti12(
            i12index1,
            i12index2,
            i12value,
            orbinit,
            orbfinal,
            args.threshold,
            s1index,
            s1value,
            order,
            zoom,
        )


if __name__ == "__main__":
    main()
