#!/usr/bin/env python

import numpy as np

from pybest.gbasis import (
    get_gobasis,
    compute_eri,
    compute_kinetic,
    compute_nuclear,
    compute_nuclear_repulsion,
    compute_overlap,
)
from pybest.io import IOData
from pybest.linalg import DenseLinalgFactory
from pybest.occ_model import AufbauOccModel
from pybest.wrappers import RHF
from pybest.units import angstrom

# Hartree-Fock calculation
# ------------------------

# Construct a molecule from scratch
bond_length = 1.098 * angstrom
mol = IOData(title="dinitrogen")
mol.coordinates = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, bond_length]])
mol.atom = np.array(["N", "N"])

# Create a Gaussian basis set
gobasis = get_gobasis("cc-pvdz", mol)

# Create a linalg factory
lf = DenseLinalgFactory(gobasis.nbasis)

# Compute integrals
olp = compute_overlap(gobasis)
kin = compute_kinetic(gobasis)
ne = compute_nuclear(gobasis)
eri = compute_eri(gobasis)
external = compute_nuclear_repulsion(gobasis)

# Create alpha orbitals
orb_a = lf.create_orbital()

# Decide how to occupy the orbitals (7 alpha electrons)
occ_model = AufbauOccModel(gobasis)

# Converge WFN
hf = RHF(lf, occ_model)
hf_output = hf(kin, ne, eri, external, olp, orb_a)

# Write SCF results to a file
hf_output.to_file("n2-scf.molden")
