# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Ionization Potential Equation of Motion Coupled Cluster implementations.
   This is a base class for all single IP implementations.

   Variables used in this module:
    :ncore:     number of frozen core orbitals
                (abbreviated as nc)
    :nocc:      number of occupied orbitals in the principle configuration
    :nacto:     number of active occupied orbitals in the principle configuration
                (abbreviated as no)
    :nvirt:     number of virtual orbitals in the principle configuration
    :nactv:     number of active virtual orbitals in the principle configuration
                (abbreviated as nv)
    :nbasis:    total number of basis functions
    :nact:      total number of active orbitals (nacto+nactv)
                (abbreviated as na)
    :alpha:     the number of unpaired electrons, determines s_z
    :e_ip:      the energy correction for IP
    :civ_ip:    the CI amplitudes from a given EOM model
    :e_ref:     the total energy of the (CC) referece wave function
    :alpha:     number of unpaired electrons

   Indexing convention:
    :i,j,k,..: occupied orbitals of principle configuration
    :a,b,c,..: virtual orbitals of principle configuration
    :p,q,r,..: general indices (occupied, virtual)
"""

import numpy as np

from pybest.exceptions import ArgumentError
from pybest.ip_eom.xip_base import RXIPCC


class RSIPCC(RXIPCC):
    """
    Restricted Ionization Potential Equation of Motion Coupled Cluster
    class restricted to single IP

    Purpose:
    Determine the excitation energies from some IP-EOM-CC model

    Currently supported particle-hole operators:
     * hole (1h)
     * hole-particle-hole (2h1p)

    """

    long_name = ""
    acronym = "IP-EOM-CC"
    reference = ""
    order = "IP"
    alpha = -1

    @RXIPCC.nhole.setter
    def nhole(self, new):
        """Set number of hole operators.

        **Arguments**

        new
            Allowed values are either in [1,2] or None. If None is passed,
            nhole is defaulted to 2 hole operators.
        """
        if new is None:
            # set default value to two hole operators
            self._nhole = 2
        elif new in [1, 2]:
            self._nhole = new
        else:
            raise NotImplementedError
        if not self._check_nhole:
            raise ArgumentError(
                "Unsupported number of hole operators and occupied orbitals. "
                f"We have {self.occ_model.nacto[0]} occupied orbitals and {self._nhole} "
                f"hole operators for {self.alpha} unpaired electrons."
            )

    @property
    def dimension(self):
        """Total number of unknowns of chosen IP model"""
        raise NotImplementedError

    def _check_nhole(self):
        """Check for consistency of the number of hole operators. Returns
        False if the number of hole operators does not agree with the spin
        projection and the number of active occupied orbitals.
        """
        return self.nhole <= self.occ_model.nacto[0]

    def set_hamiltonian(self, mo1, mo2):
        """Update auxiliary tensors for specific IP model"""
        raise NotImplementedError

    def compute_h_diag(self, *args):
        """Used by the Davidson module for pre-conditioning"""
        raise NotImplementedError

    def build_subspace_hamiltonian(self, b_vector, h_diag, *args):
        """Used by the Davidson module to construct subspace Hamiltonian"""
        raise NotImplementedError

    ###########################################################################
    # Printing operations for all IP models
    ###########################################################################

    def print_ci_vector(self, ci_dict):
        """Print eigenvectors for various S_z

        **Arguments:**

        ci_dict:
            (OrderedDict) composite index as key, CI vector as value
        """
        raise NotImplementedError

    def print_weights(self, e_vec_j):
        """Print weights of R operators for various S_z

        **Arguments:**

        e_vec_j:
            (np.array) CI vector
        """
        raise NotImplementedError

    ###########################################################################
    # Indexing operations
    ###########################################################################

    def get_mask(self, select):
        """Get unique indices that are returned during diagonalization

        **Arguments:**

        select:
            (int) diagonal offset as defined in numpy.triu_indices (k arg)
        """
        no, nv = self.occ_model.nacto[0], self.occ_model.nactv[0]
        ind0, ind1 = np.indices((no, nv))
        mask = np.ones((no, no, nv), dtype=bool)
        mask[ind0, ind0, ind1] = False
        mask[np.triu_indices(no, k=select)] = False
        return mask

    def get_index_of_mask(self, select):
        """Get unique indices of mask to assign during diagonalization

        **Arguments:**

        select:
            (int) diagonal offset as defined in numpy.triu_indices (k arg)
        """
        mask = self.get_mask(select)
        indices = np.where(mask)
        return indices

    def get_index_ijb(self, index):
        """
        Return hole-hole-particle indices from composite index of CI vector

        ** Arguments: **

        index
            The composite index to be resolved
        """
        no, nv = self.occ_model.nacto[0], self.occ_model.nactv[0]
        if self.spin_free:
            i, j, b = np.unravel_index(index, (no, no, nv))
            return i, j, b, False
        end_aa = (no - 1) * no * nv // 2
        same_spin = True
        # Full index:
        if index < end_aa:
            #
            # return ijb block
            #
            # Triu index:
            triu = np.triu_indices(no, k=1)
            b = index % nv
            index_ = index // nv
            i = triu[0][index_]
            j = triu[1][index_]
        else:
            #
            # return iJB block
            #
            i, j, b = np.unravel_index(index - end_aa, (no, no, nv))
            same_spin = False
        return i, j, b, same_spin
