# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --

# pylint: disable=too-many-arguments, no-member, protected-access

import numpy as np
import pytest

import pybest.gbasis.gobasis as pybest_basis
from pybest.context import context
from pybest.exceptions import ArgumentError
from pybest.geminals.rpccd import RpCCD, ROOpCCD
from pybest.ip_eom.xip_pccd import RDIPpCCD
from pybest.ip_eom.dip_pccd0 import RDIPpCCD0
from pybest.ip_eom.dip_pccd2 import RDIPpCCD2
from pybest.ip_eom.dip_pccd4 import RDIPpCCD4
from pybest.ip_eom.tests.common import Molecule
from pybest.linalg.dense import DenseLinalgFactory
from pybest.linalg.cholesky import CholeskyLinalgFactory
from pybest.occ_model import AufbauOccModel


testdata_class = [(RDIPpCCD0, 0), (RDIPpCCD2, 2), (RDIPpCCD4, 4)]


@pytest.mark.parametrize("cls, alpha", testdata_class)
def test_class_instance_dippccd(cls, alpha):
    """Check whether alpha keyword generates proper instance of class."""
    # some preliminaries
    lf = DenseLinalgFactory(10)
    occ_model = AufbauOccModel(lf, nel=8)
    # Initialize empty class
    ipcc = RDIPpCCD(lf, occ_model, alpha=alpha)

    assert isinstance(ipcc, cls)


#
# Test dimension
#

test_data_dim = [
    # nbasis, nhole, nocc, kwargs, expected
    (10, 2, 2, {"ncore": 0, "alpha": 0}, 4),
    (10, 3, 2, {"ncore": 0, "alpha": 0}, 36),
    (10, 2, 2, {"ncore": 0, "alpha": 2}, 1),
    (10, 3, 2, {"ncore": 0, "alpha": 2}, 17),
    (10, 3, 2, {"ncore": 0, "alpha": 4}, 0),
    (10, 2, 2, {"ncore": 1, "alpha": 0}, 1),
    (10, 3, 2, {"ncore": 1, "alpha": 0}, 1),
    (10, 2, 2, {"ncore": 1, "alpha": 2}, 0),
    (10, 3, 2, {"ncore": 1, "alpha": 2}, 0),
    (10, 3, 2, {"ncore": 1, "alpha": 4}, 0),
    (10, 2, 4, {"ncore": 0, "alpha": 0}, 16),
    (10, 3, 4, {"ncore": 0, "alpha": 0}, 304),
    (10, 2, 4, {"ncore": 0, "alpha": 2}, 6),
    (10, 3, 4, {"ncore": 0, "alpha": 2}, 174),
    (10, 3, 4, {"ncore": 0, "alpha": 4}, 24),
    (10, 2, 4, {"ncore": 1, "alpha": 0}, 9),
    (10, 3, 4, {"ncore": 1, "alpha": 0}, 117),
    (10, 2, 4, {"ncore": 1, "alpha": 2}, 3),
    (10, 3, 4, {"ncore": 1, "alpha": 2}, 63),
    (10, 3, 4, {"ncore": 1, "alpha": 4}, 6),
    (10, 2, 8, {"ncore": 0, "alpha": 0}, 64),
    (10, 3, 8, {"ncore": 0, "alpha": 0}, 960),
    (10, 2, 8, {"ncore": 0, "alpha": 2}, 28),
    (10, 3, 8, {"ncore": 0, "alpha": 2}, 588),
    (10, 3, 8, {"ncore": 0, "alpha": 4}, 112),
]


@pytest.mark.parametrize("nbasis,nhole,nocc,kwargs,expected", test_data_dim)
def test_dip_pccd_dimension(nbasis, nhole, nocc, kwargs, expected):
    """Test number of unkowns (CI coefficients) for various parameter sets
    (alpha, ncore, nocc)
    """
    # Create DIPpCCD instance
    lf = DenseLinalgFactory(nbasis)
    occ_model = AufbauOccModel(lf, nel=nocc * 2, ncore=kwargs.get("ncore"))
    dippccd = RDIPpCCD(lf, occ_model, **kwargs)
    # overwrite private attribute
    dippccd._nhole = nhole

    assert expected == dippccd.dimension


#
# Test unmask function
#

mol = Molecule("c", "cc-pvdz", DenseLinalgFactory, charge=-2)

test_unmask_error = [
    # no t_p
    (mol, {"alpha": 0}, (), ArgumentError),
    (mol, {"alpha": 2}, (), ArgumentError),
    (mol, {"alpha": 4}, (), ArgumentError),
    # no olp
    (mol, {"alpha": 0}, (mol.t_p,), ArgumentError),
    (mol, {"alpha": 2}, (mol.t_p,), ArgumentError),
    (mol, {"alpha": 4}, (mol.t_p,), ArgumentError),
    # no orb
    (mol, {"alpha": 0}, (mol.t_p, mol.olp), ArgumentError),
    (mol, {"alpha": 2}, (mol.t_p, mol.olp), ArgumentError),
    (mol, {"alpha": 4}, (mol.t_p, mol.olp), ArgumentError),
    # no two
    (mol, {"alpha": 0}, (mol.t_p, mol.olp, *mol.orb), UnboundLocalError),
    (mol, {"alpha": 2}, (mol.t_p, mol.olp, *mol.orb), UnboundLocalError),
    (mol, {"alpha": 4}, (mol.t_p, mol.olp, *mol.orb), UnboundLocalError),
]


@pytest.mark.parametrize("mol_, kwargs, args, raised_error", test_unmask_error)
def test_dip_pccd_unmask_raise_error(mol_, kwargs, args, raised_error):
    """Test unmask_arg function by passing insufficient arguments"""
    # Create EApCCD instance
    dippccd = RDIPpCCD(mol_.lf, mol_.occ_model, **kwargs)

    with pytest.raises(raised_error):
        assert dippccd.unmask_args(*args)


test_unmask = [
    # everything should be fine
    (
        mol,
        {"alpha": 0},
        (mol.t_p, mol.olp, *mol.orb, *mol.hamiltonian),
        (mol.one, mol.two, mol.orb[0]),
    ),
    (
        mol,
        {"alpha": 2},
        (mol.t_p, mol.olp, *mol.orb, *mol.hamiltonian),
        (mol.one, mol.two, mol.orb[0]),
    ),
    (
        mol,
        {"alpha": 4},
        (mol.t_p, mol.olp, *mol.orb, *mol.hamiltonian),
        (mol.one, mol.two, mol.orb[0]),
    ),
]


@pytest.mark.parametrize("mol_, kwargs, args, expected", test_unmask)
def test_ea_pccd_unmask(mol_, kwargs, args, expected):
    """Test unmask_arg function by passing proper arguments"""
    # Create DIPpCCD instance
    dippccd = RDIPpCCD(mol_.lf, mol_.occ_model, **kwargs)

    # we do not need to check the arrays, just the objects
    assert dippccd.unmask_args(*args) == expected


#
# Test effective Hamiltonian (only initialization)
#

# h_alpha_nhole:
h_0_2 = {"fock", "x1im", "goovv", "goooo"}
h_0_3 = {
    "fock",
    "x1im",
    "goovv",
    "goooo",
    "x1cd",
    "gooov",
    "xjkcm",
    "xkcmd",
    "xkcMD",
    "xjcmD",
    "xjkmN",
    "xikcm",
}
h_2_2 = {"fock", "x1im", "goooo"}
h_2_3 = {
    "fock",
    "x1im",
    "goooo",
    "goovv",
    "x1cd",
    "gooov",
    "xjkcm",
    "xkcmd",
    "xkcMD",
    "xjcmD",
    "xjkmN",
}
h_4_3 = {"fock", "x1im", "x1cd", "goooo", "x2kcmd"}

test_set_hamiltonian = [
    # molecule instance, nparticle, alpha, expected
    (mol, 2, {"alpha": 0}, h_0_2),
    (mol, 2, {"alpha": 2}, h_2_2),
    (mol, 3, {"alpha": 0}, h_0_3),
    (mol, 3, {"alpha": 2}, h_2_3),
    (mol, 3, {"alpha": 4}, h_4_3),
]


@pytest.mark.parametrize("mol_,nhole,kwargs,expected", test_set_hamiltonian)
def test_dip_pccd_set_hamiltonian(mol_, nhole, kwargs, expected):
    """Test if effective Hamiltonian has been constructed at all. We do not
    test the actual elements.
    """
    # Create DIPpCCD instance
    dippccd = RDIPpCCD(mol_.lf, mol_.occ_model, **kwargs)
    # set some class attributes
    dippccd.unmask_args(mol_.t_p, mol_.olp, *mol_.orb, *mol_.hamiltonian)
    dippccd._nhole = nhole

    # we do not need to check the arrays, just the objects
    # we need to copy the arrays as they get deleted
    one, two = mol_.one.copy(), mol_.two.copy()
    dippccd.set_hamiltonian(mol_.one, mol_.two)
    mol_.one, mol_.two = one, two

    # Check if cache instance contains all relevant terms
    assert dippccd.cache._store.keys() == expected, "Cache element not found"
    # Check loading from cache
    for h_eff in expected:
        assert dippccd.from_cache(h_eff), "Loading from cache unsuccesful"


if pybest_basis.PYBEST_CHOLESKY_ENABLED:
    test_linalg_set = [DenseLinalgFactory, CholeskyLinalgFactory]
else:
    test_linalg_set = [DenseLinalgFactory]

test_data_dip = [
    (
        RpCCD,
        "c",
        "cc-pvdz",
        {"ncore": 0, "charge": -2, "alpha": 0, "nroot": 4, "nhole": 3},
        {},
        {
            "e_tot": -37.181350747097,
            "e_ip_0": [
                -5.403875e-01,
                -4.853248e-01,
                -4.851959e-01,
                -4.643491e-01,
            ],
            "orbfile": "c_rhf.txt",
        },
    ),
    (
        RpCCD,
        "c",
        "cc-pvdz",
        {"ncore": 0, "charge": -2, "alpha": 2, "nroot": 4, "nhole": 3},
        {},
        {
            "e_tot": -37.181350747097,
            "e_ip_2": [
                -5.403875e-01,
                -4.643491e-01,
                -4.643491e-01,
                -2.845116e-01,
            ],
            "orbfile": "c_rhf.txt",
        },
    ),
    (
        RpCCD,
        "c",
        "cc-pvdz",
        {"ncore": 0, "charge": -2, "alpha": 4, "nroot": 4, "nhole": 3},
        {},
        {
            "e_tot": -37.181350747097,
            "e_ip_4": [
                -2.845116e-01,
                3.529028e-01,
                3.529028e-01,
                4.781519e-01,
                4.961400e-01,
            ],
            "orbfile": "c_rhf.txt",
        },
    ),
    (
        RpCCD,
        "c",
        "cc-pvdz",
        {"ncore": 1, "charge": -2, "alpha": 0, "nroot": 4, "nhole": 3},
        {},
        {
            "e_tot": -37.181056931067,
            "e_ip_0": [
                -5.401817e-01,
                -4.851401e-01,
                -4.850112e-01,
                -4.643028e-01,
            ],
            "orbfile": "c_rhf.txt",
        },
    ),
    (
        RpCCD,
        "c",
        "cc-pvdz",
        {"ncore": 1, "charge": -2, "alpha": 2, "nroot": 4, "nhole": 3},
        {},
        {
            "e_tot": -37.181056931067,
            "e_ip_2": [
                -5.401817e-01,
                -4.643028e-01,
                -4.643028e-01,
                -2.844379e-01,
            ],
            "orbfile": "c_rhf.txt",
        },
    ),
    (
        ROOpCCD,
        "c",
        "cc-pvdz",
        {"ncore": 0, "charge": -2, "alpha": 0, "nroot": 4, "nhole": 3},
        {"sort": False},
        {
            "e_tot": -37.194223264348,
            "e_ip_0": [
                -5.264451e-01,
                -4.749906e-01,
                -4.538110e-01,
                -4.538110e-01,
            ],
            "orbfile": "c_oopccd.txt",
        },
    ),
    (
        ROOpCCD,
        "c",
        "cc-pvdz",
        {"ncore": 0, "charge": -2, "alpha": 2, "nroot": 4, "nhole": 3},
        {"sort": False},
        {
            "e_tot": -37.194223264348,
            "e_ip_2": [
                -5.264451e-01,
                -4.538110e-01,
                -4.538110e-01,
                -2.749086e-01,
            ],
            "orbfile": "c_oopccd.txt",
        },
    ),
    (
        ROOpCCD,
        "c",
        "cc-pvdz",
        {"ncore": 0, "charge": -2, "alpha": 4, "nroot": 5, "nhole": 3},
        {"sort": False},
        {
            "e_tot": -37.194223264348,
            "e_ip_4": [
                -2.749087e-01,
                3.644533e-01,
                3.644533e-01,
                4.870770e-01,
                5.050943e-01,
            ],
            "orbfile": "c_oopccd.txt",
        },
    ),
]


@pytest.mark.parametrize(
    "cls, mol_f, basis, kwargs, kwargs_pccd, expected", test_data_dip
)
@pytest.mark.parametrize("linalg_set", test_linalg_set)
def test_dip_pccd(
    cls, mol_f, basis, kwargs, kwargs_pccd, expected, linalg_set
):
    """Test attachement energies of DIPpCCD models"""
    ncore = kwargs.get("ncore")
    charge = kwargs.get("charge")
    alpha = kwargs.get("alpha")
    nroot = kwargs.get("nroot")
    nhole = kwargs.get("nhole")
    # Prepare molecule
    mol_ = Molecule(mol_f, basis, linalg_set, ncore=ncore, charge=charge)
    # Do RHF optimization:
    mol_.do_rhf()
    # overwrite orbitals (pCCD convergence speed up)
    fn_orb = context.get_fn(f"test/{expected['orbfile']}")
    mol_.hf.orb_a.coeffs[:] = np.fromfile(fn_orb, sep=",").reshape(
        mol_.obasis.nbasis, mol_.obasis.nbasis
    )
    # Do pCCD optimization:
    mol_.do_pccd(cls, **kwargs_pccd)
    assert abs(mol_.pccd.e_tot - expected["e_tot"]) < 1e-6
    # Do DIPpCCD optimization:
    mol_.do_dip_pccd(alpha, nroot, nhole)

    for ind in range(nroot):
        # get proper key (e_ip_0 or e_ip_2 or e_ip_4)
        (key,) = [key for key in expected if "e_ip" in key.lower()]
        assert (
            abs(getattr(mol_.dip_pccd, key)[ind] - expected[key][ind]) < 5e-6
        )
