# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Tipm
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --

# pylint: disable=too-many-arguments

import pytest
import numpy as np

from pybest.ip_eom.sip_base import RSIPCC
from pybest.ip_eom.sip_pccd1 import RIPpCCD1
from pybest.ip_eom.dip_base import RDIPCC
from pybest.linalg.dense import DenseLinalgFactory, DenseOneIndex
from pybest.occ_model import AufbauOccModel

test_data_range = [
    (
        "oovv",
        0,
        10,
        4,
        0,
        {
            "begin0": 0,
            "end0": 4,
            "begin1": 0,
            "end1": 4,
            "begin2": 4,
            "end2": 10,
            "begin3": 4,
            "end3": 10,
        },
    ),
    ("oV", 0, 10, 4, 0, {"begin0": 0, "end0": 4, "begin1": 0, "end1": 6}),
    ("nn", 0, 10, 4, 0, {"begin0": 0, "end0": 10, "begin1": 0, "end1": 10}),
    ("nn", 0, 10, 4, 1, {"begin0": 0, "end0": 9, "begin1": 0, "end1": 9}),
    ("ov", 2, 10, 4, 0, {"begin2": 0, "end2": 4, "begin3": 4, "end3": 10}),
    ("ov", 2, 10, 4, 1, {"begin2": 0, "end2": 3, "begin3": 3, "end3": 9}),
]

test_ip_cls = [RSIPCC, RDIPCC]


@pytest.mark.parametrize(
    "range_,start,nbasis,nocc,ncore,expected", test_data_range
)
@pytest.mark.parametrize("cls", test_ip_cls)
def test_get_range(range_, start, nbasis, nocc, ncore, expected, cls):
    """Test if proper ranges are selected when cache is initialized."""
    # some preliminaries
    lf = DenseLinalgFactory(nbasis)
    occ_model = AufbauOccModel(lf, nel=nocc * 2, ncore=ncore)
    # Initialize empty class (we cannot use abc class)
    ipcc = cls(lf, occ_model)

    assert ipcc.get_range(range_, start) == expected


test_data_size = [
    # function argument, expected
    ("ovv", 10, 4, 0, (4, 6, 6)),
    ("ovv", 10, 4, 2, (2, 6, 6)),
    ("ooov", 10, 4, 0, (4, 4, 4, 6)),
    ("ooov", 10, 4, 3, (1, 1, 1, 6)),
    ("ooovvv", 10, 4, 3, (1, 1, 1, 6, 6, 6)),
    ((3, 5, 19), 10, 4, 0, (3, 5, 19)),
    ((3, 5, 19), 10, 4, 2, (3, 5, 19)),
    ((3,), 10, 4, 2, (3,)),
    ((3, 3, 6, 6, 9), 10, 4, 2, (3, 3, 6, 6, 9)),
]


@pytest.mark.parametrize("size_,nbasis,nocc,ncore,expected", test_data_size)
@pytest.mark.parametrize("cls", test_ip_cls)
def test_get_size(size_, nbasis, nocc, ncore, expected, cls):
    """Test if proper sizes are returned when cache is initialized."""
    # some preliminaries
    lf = DenseLinalgFactory(nbasis)
    occ_model = AufbauOccModel(lf, nel=nocc * 2, ncore=ncore)
    # Initialize empty class (we cannot use abc class)
    ipcc = cls(lf, occ_model)

    assert ipcc.get_size(size_) == expected


h_diag_ = DenseOneIndex(10)
h_diag_.label = "h_diag"
h_diag_.array[:] = np.array([1, 4, 2, 6, 8, 9, 4, 5, 7, 3])
vec1 = DenseOneIndex(10)
vec2 = DenseOneIndex(10)
vec3 = DenseOneIndex(10)
vec4 = DenseOneIndex(10)
vec1.set_element(0, 1.0)
vec2.set_element(2, 1.0)
vec3.set_element(9, 1.0)
vec4.set_element(1, 1.0)

test_data_guess = [
    # nguess, h_diag, expected
    (1, h_diag_, ([vec1], 1)),
    (2, h_diag_, ([vec1, vec2], 2)),
    (3, h_diag_, ([vec1, vec2, vec3], 3)),
    (4, h_diag_, ([vec1, vec2, vec3, vec4], 4)),
]


@pytest.mark.parametrize("nguess,h_diag,expected", test_data_guess)
def test_build_guess_vectors(nguess, h_diag, expected):
    """Test if proper guess vectors are constructed. We only test for the
    ippCCD1 class and 1 particle operator (ipsiest as we have only r_a)
    as we need the dimension property attribute to be defined correctly.
    All other classes use the same function.
    """
    # some preliminaries, we always test for 11 basis functions and 10 nacto
    lf = DenseLinalgFactory(11)
    occ_model = AufbauOccModel(lf, nel=20)
    # Initialize empty class (we cannot use abc class)
    ipcc = RIPpCCD1(lf, occ_model)
    # Write property attribute
    ipcc.nhole = 1

    guessv, nguessv = ipcc.build_guess_vectors(nguess, False, h_diag)
    # check length
    assert nguessv == expected[1], "wrong length of guess vector"
    # check ipch vector
    for vec_t, vec_r in zip(guessv, expected[0]):
        assert (
            vec_t.array == vec_r.array
        ).all(), "wrong elements in guess vector"


test_data_ci_sort = [
    # h_diag_ = np.array([1,4,2,6,8,9,4,5,7,3])
    (10, 8, h_diag_.array, {5: 9}),
    (10, 8, -h_diag_.array, {5: -9}),
    (10, 3.9, h_diag_.array - 5, {0: -4, 5: 4}),
    (10, 7.9, h_diag_.array, {5: 9, 4: 8}),
    (10, 7.9, -h_diag_.array, {5: -9, 4: -8}),
    (10, 3.9, h_diag_.array, {5: 9, 4: 8, 8: 7, 3: 6, 7: 5, 1: 4, 6: 4}),
]


@pytest.mark.parametrize("nbasis,threshold,ci_v,expected", test_data_ci_sort)
def test_sort_ci_vectors(nbasis, threshold, ci_v, expected):
    """Test if CI vectors are properly sorted."""
    # some preliminaries
    lf = DenseLinalgFactory(nbasis)
    # It does not matter what we set here
    occ_model = AufbauOccModel(lf, nel=2)
    # Initialize empty class (we cannot use abc class)
    ipcc = RIPpCCD1(lf, occ_model)

    ci_ordered = ipcc.sort_ci_vector(ci_v, threshold)
    for d_t, d_r in zip(ci_ordered.items(), expected.items()):
        assert d_t == d_r, "sorting did not work properly"
